#!/usr/bin/env node

/* eslint no-console:0, no-param-reassign:0 */

const meow = require('meow')
const updateNotifier = require('update-notifier')
const fs = require('fs')
const objectPath = require('object-path')
const globby = require('globby')
const jsonfile = require('jsonfile')
const boxen = require('boxen')

const hasOwnProperty = (what, property) => Object.prototype.hasOwnProperty.call(what, property)

const IGNORES = [
  'node_modules',
  'logs',
  'pids',
  'coverage',
  'bower_components',
  'jspm_packages',
  'typings',
]
const cli = meow(`
  Usage
    $ bitsausage

  Options
    --force, -f  Will create .bithoundrc even if BitHound is not in README

  Example
    Just call it in the root of your package
`, {
    alias: {
      f: 'force',
    },
  })
updateNotifier({ pkg: cli.pkg }).notify()

function sortObject(obj) {
  return Object.keys(obj).sort().reduce((result, key) => {
    result[key] = obj[key]
    return result
  }, {})
}

function aContainsB(a, b) {
  return a.indexOf(b) >= 0
}

// step 4.
function step4(bithoundrc) {
  jsonfile.writeFile('.bithoundrc', bithoundrc, { spaces: 2 }, (err) => {
    if (err) {
      console.error(`BitSausage: ${err}`)
      process.exit(1)
    }
    process.exit(0)
  })
}

// step 3.
function step3(bithoundrc) {
  // find test files and generate entries in .bithoundrc for them
  const testGlob = globby.sync(['**/*test*.js'].concat(IGNORES.map(el => `!**${el}/**`)))
  if (Array.isArray(testGlob) && testGlob.length > 0) {
    objectPath.set(bithoundrc, 'test', testGlob)
  }
  step4(sortObject(bithoundrc))
}

// step 2.
function step2(bithoundrc) {
  fs.readFile('./package.json', 'utf8', (packageErr, packageData) => {
    let blanketDeps = []
    if (packageData) {
      // PART I. DEPS WITH BLANKET RANGES ("*")
      packageData = JSON.parse(packageData)
      if (hasOwnProperty(packageData, 'dependencies') && Object.keys(packageData.dependencies).length > 0) {
        blanketDeps = blanketDeps.concat(Object.keys(packageData.dependencies)
          .filter(dep => packageData.dependencies[dep] === '*'))
      }
      if (hasOwnProperty(packageData, 'devDependencies') && Object.keys(packageData.devDependencies).length > 0) {
        blanketDeps = blanketDeps.concat(Object.keys(packageData.devDependencies)
          .filter(dep => packageData.devDependencies[dep] === '*'))
      }
      if (blanketDeps.length > 0) {
        objectPath.set(bithoundrc, 'dependencies.mute', blanketDeps.sort())
      }

      // PART II. CRITICS (LINTING DETECTION)
      if (hasOwnProperty(packageData, 'devDependencies')) {
        // BitHound's supported linters:
        // * jshint
        // * eslint
        // * tslint
        // * standard
        // * semistandard (because some of us like semicolons)
        // * none (in case you want to disable lint)
        // source: http://blog.bithound.io/configuring-bithound-101/
        // retrieved: 2017-06-19

        if (hasOwnProperty(packageData.devDependencies, 'standard')) {
          objectPath.set(bithoundrc, 'critics.lint.engine', 'standard')
        } else if (hasOwnProperty(packageData.devDependencies, 'jshint')) {
          objectPath.set(bithoundrc, 'critics.lint.engine', 'jshint')
        } else if (hasOwnProperty(packageData.devDependencies, 'eslint')) {
          objectPath.set(bithoundrc, 'critics.lint.engine', 'eslint')
        } else if (hasOwnProperty(packageData.devDependencies, 'tslint')) {
          objectPath.set(bithoundrc, 'critics.lint.engine', 'tslint')
        } else if (hasOwnProperty(packageData.devDependencies, 'semistandard')) {
          objectPath.set(bithoundrc, 'critics.lint.engine', 'semistandard')
        } else {
          console.log(boxen('Strange!\nYour project has no linting set.\nWe\'ll turn off bitHound linting.', { padding: 1, borderColor: 'red' }))
          objectPath.set(bithoundrc, 'critics.lint.engine', 'none')
        }
      } else {
        objectPath.set(bithoundrc, 'critics.lint.engine', 'none')
      }

      // time for interlude

      //                  ________________________
      //       /\___/\   /                        \
      //      ( o   o )  | mmm... cod fillets...  )
      //      (  =^=  ) <________________________/
      //      (        )
      //      (         )
      //      (          )))))))))))
      //


      // PART III. ADD BABEL OUTPUT AS IGNORES
      // both files and folders, recognise both Babel's --out-dir and --out-file

      // Let's not be cheeky and assume that npm scripts containing Babel builds
      // will always be under key "build". Let's do this properly and iterate all
      // npm scripts.

      objectPath.set(bithoundrc, 'ignore', [])
      let babelBuildScriptRow = null
      if (packageData.scripts) {
        Object.keys(packageData.scripts).forEach((oneOfScripts) => {
          if (
            packageData.scripts[oneOfScripts].includes('babel') &&
            (
              packageData.scripts[oneOfScripts].includes('--out-dir') ||
              packageData.scripts[oneOfScripts].includes('--out-file')
            )
          ) {
            babelBuildScriptRow = packageData.scripts[oneOfScripts]
          }
        })
      }

      if (babelBuildScriptRow) {
        babelBuildScriptRow
          .split(' ')
          .filter(el => el.length > 0)
          .forEach((chunkBetweenSpaces, index, wholeArray) => {
            // prep --out-file
            if (
              (chunkBetweenSpaces === '--out-file') &&
              (wholeArray[index + 1] !== undefined)
            ) {
              bithoundrc.ignore.push(wholeArray[index + 1])
            }
            // prep --out-dir
            if (
              (chunkBetweenSpaces === '--out-dir') &&
              (wholeArray[index + 1] !== undefined)
            ) {
              bithoundrc.ignore.push(`${wholeArray[index + 1]}/**`)
            }
          })
      }

      // PART IV. ADD DIST FOLDER TO IGNORES LIST (IF SUCH EXISTS)
      if (fs.existsSync('dist')) {
        bithoundrc.ignore.push('dist/**')
      }

      // FIN.
      step3(bithoundrc)
    } else if (packageErr) {
      step3(bithoundrc)
    }
  })
}

// step 1.
fs.readFile('./readme.md', 'utf8', (readmeErr, readmeData) => {
  if (
    cli.flags.force || (
      readmeData &&
      (typeof readmeData === 'string') &&
      (readmeData.length > 0) &&
      aContainsB(readmeData.toLowerCase(), 'bithound')
    )
  ) {
    step2({})
  }
})
